#!/usr/bin/python

from lpltk import LaunchpadService

try:
    from lazr.restfulclient.errors import NotFound as nf_error
except:
    from lazr.restfulclient.errors import HTTPError as nf_error

from datetime import datetime, timedelta
import optparse
import simplejson as json
import sys


def process_tasks(recent_tasks, end_date=None):
    apport_tags = ['apport-crash', 'apport-bug', 'apport-package',
        'apport-lpi', 'apport-kerneloops']
    packages = {}
    for task in recent_tasks:
        if end_date:
            # because launchpadlib doesn't return a datetime any more for
            # date_created LP: #924240
            if isinstance(task.date_created, unicode):
                dc = datetime.strptime(task.date_created.replace('+00:00',''),
                    '%Y-%m-%dT%H:%M:%S.%f')
            else:
                dc = task.date_created.replace(tzinfo=None)
            if dc > end_date:
                continue
        src_pkg = task.bug_target_name.split(' ')[0]
        if src_pkg == 'ubuntu':
            continue
        if src_pkg not in packages:
            packages[src_pkg] = {}
        packages[src_pkg]['total'] = packages[src_pkg].setdefault('total', 0) + 1
        bug = task.bug
        if bug.duplicate_of:
            try:
                master = bug.duplicate_of.id
            except nf_error as error:
                master = error.content.split(':')[-1].strip(" u'")
            packages[src_pkg]['dupes'] = packages[src_pkg].setdefault('dupes', 0) + 1
            if 'masters' not in packages[src_pkg]:
                packages[src_pkg]['masters'] = {}
            packages[src_pkg]['masters'][master] = packages[src_pkg]['masters'].setdefault(master, 0) + 1
        tags = bug.tags
        if not set(apport_tags).intersection(tags):
            packages[src_pkg]['other'] = packages[src_pkg].setdefault('other', 0) + 1
            continue
        # apport-lpi bugs also get tagged apport-bug and we
        # don't want to count bug tasks 2 times so bail out
        if 'apport-lpi' in tags:
            packages[src_pkg]['apport-lpi'] = packages[src_pkg].setdefault('apport-lpi', 0) + 1
            continue
        else:
            for apport_tag in apport_tags:
                if apport_tag in tags:
                    packages[src_pkg]['%s' % apport_tag] = packages[src_pkg].setdefault(apport_tag, 0) + 1
    # In the event that none of the bug tasks has an apport_tag set the value
    # to zero
    for package in packages:
        for apport_tag in apport_tags:
            if apport_tag not in packages[package]:
                packages[package][apport_tag] = 0
        if 'other' not in packages[package]:
            packages[package]['other'] = 0
        if 'dupes' not in packages[package]:
            packages[package]['dupes'] = 0
        if 'masters' not in packages[package]:
            packages[package]['masters'] = {}
    return packages


def find_bug_tasks(start_date, team=None, package=None):
    if team:
        recent_tasks = ubuntu.searchTasks(created_since=start_date,
            status=['New', 'Incomplete', 'Incomplete (with response)',
                'Incomplete (without response)', 'Confirmed', 'Triaged',
                'In Progress', 'Fix Committed', 'Fix Released', 'Invalid',
                'Opinion', "Won't Fix"],
            bug_supervisor=team, order_by='targetname', omit_duplicates=False)
    elif package:
        recent_tasks = package.searchTasks(created_since=start_date,
            status=['New', 'Incomplete', 'Incomplete (with response)',
                'Incomplete (without response)', 'Confirmed', 'Triaged',
                'In Progress', 'Fix Committed', 'Fix Released', 'Invalid',
                'Opinion', "Won't Fix"], order_by='targetname')
    return recent_tasks


def prepare_data(package_frequency, period):
    period_data = []

    pkg_totals = {}
    for pkg in package_frequency:
        pkg_totals[pkg] = package_frequency[pkg]['total']

    sorted_totals = sorted(pkg_totals, reverse=True)

    for pkg in sorted_totals:
        # remove dashes from tag names for easier javascript parsing
        period_data.append({'package': pkg,
            'total': package_frequency[pkg]['total'],
            'apportcrash': package_frequency[pkg]['apport-crash'],
            'apportbug': package_frequency[pkg]['apport-bug'],
            'apportlpi': package_frequency[pkg]['apport-lpi'],
            'apportpackage': package_frequency[pkg]['apport-package'],
            'apportkerneloops': package_frequency[pkg]['apport-kerneloops'],
            'dupes': package_frequency[pkg]['dupes'],
            'masters': package_frequency[pkg]['masters'],
            'other': package_frequency[pkg]['other'],
            'period': period})

    return period_data


def create_json_file(filename, data):
    json_file = open(filename, 'w')
    json_file.write(json.dumps(data, indent=4))
    json_file.close()


if __name__ == '__main__':
    lp = LaunchpadService(config=
        {'read_only': True,
         'display_name': 'foundations-recent-package-bug-tasks'})
    ubuntu = lp.load_project('ubuntu')

    parser = optparse.OptionParser()
    parser.add_option("--period", help="Gather bug tasks reported \
        within X day(s) in two intervals")
    parser.add_option("--team", help="Team name")
    parser.add_option("--packagesets", default=False, action="store_true")

    (opt, args) = parser.parse_args()

    if opt.team and opt.packagesets:
        print("One must choose either packagesets or team, not both")
        sys.exit(1)

    today = datetime.utcnow()
    run_time = datetime.strftime(today, '%Y-%m-%d %H:%M')

    shortperiod = int(opt.period)
    longperiod = shortperiod * 2

    start_date = today - timedelta(shortperiod)
    other_target_date = today - timedelta(longperiod)

    if opt.team:
        data = []
        team = lp.launchpad.people[opt.team]

        json_filename = "%s-recent-package-bugs.json" % (opt.team)

        recent_tasks = find_bug_tasks(start_date, team=team)
        short_pkg_freq = process_tasks(recent_tasks)
        short_period_data = prepare_data(short_pkg_freq, shortperiod)
        data.extend(short_period_data)

        tasks = find_bug_tasks(other_target_date, team=team)
        pkg_freq = process_tasks(tasks, start_date)
        long_period_data = prepare_data(pkg_freq, longperiod)
        data.extend(long_period_data)

        data.append({'ran': run_time})
        create_json_file(json_filename, data)

    if opt.packagesets:
        dev_release = ubuntu.current_series
        package_sets = [ps.name for ps in lp.launchpad.packagesets if ps.distroseries == dev_release]
        for package_set in package_sets:
            data = []
            json_filename = "%s-recent-package-bugs.json" % (package_set)
            packages = lp.launchpad.packagesets.getByName(distroseries=dev_release,
                name=package_set).getSourcesIncluded()
            for pkg_name in packages:
                src_pkg = ubuntu.getSourcePackage(name=pkg_name)
                recent_tasks = find_bug_tasks(start_date, package=src_pkg)
                short_pkg_freq = process_tasks(recent_tasks)
                short_period_data = prepare_data(short_pkg_freq, shortperiod)
                data.extend(short_period_data)

                tasks = find_bug_tasks(other_target_date, package=src_pkg)
                pkg_freq = process_tasks(tasks, start_date)
                long_period_data = prepare_data(pkg_freq, longperiod)
                data.extend(long_period_data)

            data.append({'ran': run_time})
            create_json_file(json_filename, data)
