#!/usr/bin/python3
# Author: Adam Gandelman <adamg@canonical.com>
# Copyright (C) 2012 Canonical, Ltd.
# License: GPLv3

# Using version data gathered by gather-versions.py, either generate an HTML
# report or check for outdated packages in the CA.

import common

from collections import OrderedDict
import datetime
import jinja2
import optparse
import logging
import subprocess
import json
import os

UNAVAILABLE_VER = "0~0~unavailable"

logging.basicConfig(level=logging.INFO)


def _dpkg_comp_pkg(newer, older):
    """Compare package versions."""
    if newer is None:
        newer = UNAVAILABLE_VER
    if older is None:
        older = UNAVAILABLE_VER
    dpkg = subprocess.Popen(['dpkg', '--compare-versions', newer, 'gt', older])
    dpkg.communicate()
    rc = dpkg.returncode
    if rc not in [0, 1]:
        raise Exception("'dpkg --compare-versions' returned rc other "
                        "than 0, 1")
    return rc == 0


def load_package_maps(json_dir, release, maps=None, include_upstream=False):
    """Generate package mappings for each pocket."""
    if maps is None:
        if include_upstream:
            pockets = ['ubuntu', 'staging', 'proposed', 'updates', 'upstream']
        else:
            pockets = ['ubuntu', 'staging', 'proposed', 'updates']
    i = 1
    package_maps = {}
    for pocket in pockets:
        fn = '%s_%s_versions' % (release, pocket)
        f = open(os.path.join(json_dir, fn), 'r')
        package_maps[i] = {
            'name': pocket,
            'packages': json.loads(f.read())
        }
        i += 1
    return package_maps


def render_file(search_path, template_name, render_data, out_file):
    """Render a jinja2 template out to a file."""
    loader = jinja2.FileSystemLoader(searchpath=search_path)
    jenv = jinja2.Environment(loader=loader)
    template = jenv.get_template(template_name)
    out_data = template.render(render_data=render_data)
    out = open(out_file, 'w', encoding="utf-8")
    out.write(out_data)
    out.close()


def get_template_path(subpath=''):
    """Get the path to the templates directory."""
    return os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        'templates', subpath)


def get_render_dest_path(opts, subpath, render_file):
    """Get the destination path for the rendered file."""
    if opts.out_file:
        render_dest = os.path.dirname(opts.out_file)
        render_dest = os.path.join(render_dest, subpath)
    else:
        render_dest = subpath
    if render_dest and not os.path.isdir(render_dest):
        os.mkdir(render_dest)
    return os.path.join(render_dest, render_file)


def render_css(opts):
    """Render the css files."""
    css_files = ['filter.css', 'starter-template.css', 'tracker.css']
    for css_file in css_files:
        render_file(get_template_path('css'),
                    css_file,
                    {},
                    get_render_dest_path(opts, 'css', css_file))


def render_js(opts):
    """Render the js file."""
    js_file = 'filter.js'
    render_file(get_template_path('js'),
                js_file,
                {},
                get_render_dest_path(opts, 'js', js_file))


def _newest_vers(versions):
    """Determine the newerst version in the list."""
    newest = UNAVAILABLE_VER
    for v in versions:
        if _dpkg_comp_pkg(v, newest):
            newest = v
    return newest


def _render_row(package, pkg_data, package_maps, vers_suffix):
    """Render row for package."""
    pkg_data[package] = {}

    for i in package_maps.keys():
        r = {}
        pocket = package_maps[i]['name']
        try:
            original_version = (
                package_maps[i]['packages'][package]['original_version'])
            normalized_version = (
                package_maps[i]['packages'][package]['normalized_version'])
        except KeyError:
            r.update({
                'normalized_version': '',
                'original_version': '',
                'build_state': '',
                'status': 'outdated',
            })
            pkg_data[package][pocket] = r
            continue

        if isinstance(normalized_version, list):
            # staging PPA packages also have an associated build state
            original_version, build_state = original_version
            normalized_version, build_state = normalized_version
        else:
            build_state = 'ok'

        r['build_state'] = build_state
        r['original_version'] = original_version
        r['normalized_version'] = normalized_version

        if build_state != 'ok':
            r['status'] = 'failed'

        pkg_data[package][pocket] = r

    # every cell is green if every version is the same
    if len(set([v['normalized_version']
               for k, v in pkg_data[package].items()])) == 1:
        for k, v in pkg_data[package].items():
            v['status'] = 'current'
        return

    # cells that contain the most recent version are set to 'current'
    # cells with outdated versions are set to 'outdated'
    # cells with build failures are set to 'failed'
    newest_vers = _newest_vers([v['normalized_version']
                               for k, v in pkg_data[package].items()])
    for k, v in pkg_data[package].items():
        # strip ~cloud tag so we can just use native python comparison here
        if v['build_state'] != 'ok':
            v['original_version'] += (' <small><b>({})</b></small>'.format(
                                      v['build_state']))
            v['status'] = 'failed'
        elif not v['original_version']:
            v['status'] = 'outdated'
        elif (v['normalized_version'].split(vers_suffix)[0] ==
              newest_vers.split(vers_suffix)[0]):
            v['status'] = 'current'
        else:
            v['status'] = 'outdated'

    return


def render_html(package_maps, opts):
    """Render html files for version report."""
    pkg_data = OrderedDict()
    release = opts.os_release
    render_data = {'release': release,
                   'timestamp': datetime.datetime.utcnow(),
                   'include_upstream': opts.include_upstream,
                   'pkg_data': pkg_data}
    if release == 'cloud-tools':
        render_data['release'] = 'Cloud Tools'
        vers_suffix = '~ctools'
    elif release == 'cloud-tools-next':
        render_data['release'] = 'Cloud Tools Next'
        vers_suffix = '~ctools'
    else:
        render_data['release'] = (
            'OpenStack {}{}'.format(release[0].capitalize(), release[1:])
        )
        vers_suffix = '~cloud'

    seen = []
    for i in package_maps.keys():
        for p in sorted(package_maps[i]['packages']):
            if p not in seen:
                _render_row(p, pkg_data, package_maps, vers_suffix)
                seen.append(p)

    href_map = OrderedDict()
    out_index_file = get_render_dest_path(opts, '', 'index.html')
    out_dir = os.path.dirname(os.path.realpath(out_index_file))

    hrefs = [f for f in os.listdir(out_dir)
             if os.path.isfile(os.path.join(out_dir, f)) and 'versions' in f]
    hrefs.sort()
    for href in hrefs:
        raw_name = href.split('_', 1)[0]
        if (raw_name not in common.SUPPORTED_OS_RELEASES and
                raw_name not in common.DEV_OS_RELEASES):
            continue
        name = '{}{}'.format(raw_name[0].capitalize(), raw_name[1:])
        if 'upstream' in href:
            name = '{}-{}'.format(name, 'Upstream')
        href_map[name] = {}
        href_map[name]['href'] = href
        href_map[name]['status'] = ''
        if raw_name == release:
            if opts.include_upstream and 'upstream' in href:
                href_map[name]['status'] = 'active'
            elif not opts.include_upstream and 'upstream' not in href:
                href_map[name]['status'] = 'active'
    render_data['href_map'] = href_map

    if not opts.out_file:
        if opts.include_upstream:
            out_versions_file = 'ca_upstream_versions_{}.html'.format(
                opts.os_release)
        else:
            out_versions_file = 'ca_versions_{}.html'.format(
                opts.os_release)
    else:
        out_versions_file = opts.out_file

    render_file(get_template_path(), 'versions.html.j2', render_data,
                out_versions_file)

    # ensure none of the version links are 'active' on Home page
    for name in href_map.keys():
        href_map[name]['status'] = ''
    render_file(get_template_path(), 'index.html.j2', render_data,
                out_index_file)

    return


def main():
    usage = 'usage: %prog [options]'
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-d', '--json_dir', dest='json_dir', action='store',
                      default=os.path.join(os.getenv('HOME'), '.ca-versions'),
                      help='Directory where version json data was dumped from'
                           'gather-versions.py (Default: $HOME/.ca-versions).'
                      )
    parser.add_option('-o', '--output', dest='out_file',
                      action='store', default=None,
                      help='HTML output file name.')
    parser.add_option('-r', '--os_release', dest='os_release',
                      action='store', default='folsom',
                      help='OpenStack release to track',
                      choices=sorted(common.VERSION_MAP.keys()))
    parser.add_option('-c', '--check_ubuntu', dest='check_ubuntu',
                      action='store_true', default=False,
                      help='Ensure Ubuntu pockets do not contain newer '
                           'packages than what is at least in CA staging.')
    parser.add_option('-u', '--upstream', dest='include_upstream',
                      action='store_true', default=False,
                      help='Include upstream version column in output.')
    (opts, args) = parser.parse_args()

    if not os.path.exists(opts.json_dir):
        logging.error('Directory not found: %s' % opts.json_dir)
        exit(1)

    if opts.os_release == "cloud-tools-next":
        package_maps = load_package_maps(opts.json_dir,
                                         opts.os_release,
                                         ['ubuntu', 'staging'])
        for f in package_maps.values():
            if f['name'] == "staging":
                f['header'] = "next"
    else:
        package_maps = load_package_maps(
            opts.json_dir,
            opts.os_release,
            include_upstream=opts.include_upstream)

    if opts.check_ubuntu:
        # Ensure there are no newer packages in the ubuntu release than in
        # the staging PPA.
        for p_map in package_maps:
            if package_maps[p_map]['name'] == 'staging':
                staging = package_maps[p_map]['packages']
            if package_maps[p_map]['name'] == 'ubuntu':
                ubuntu = package_maps[p_map]['packages']

        if opts.os_release.startswith('cloud-tools'):
            suffix = '~ctools'
        else:
            suffix = '~cloud'

        out = ''
        for package in staging:
            s_vers, status = staging[package]['original_version']
            s_vers = s_vers.split(suffix)[0]
            u_vers = ubuntu[package]['original_version']
            if _dpkg_comp_pkg(u_vers, s_vers):
                out += '%s:\n' % package
                out += '\tUbuntu: %s\n' % u_vers
                out += '\tCloud Archive staging: %s\n' % s_vers

        if out != '':
            print('-------------------------------------------------------')
            print('The following Cloud Archive packages for {}'.format(
                  opts.os_release))
            print('have been superseded newer versions in Ubuntu!')
            print('')
            print('{}'.format(out))
            print('-------------------------------------------------------')
            exit(1)

    render_css(opts)
    render_js(opts)
    render_html(package_maps, opts)

if __name__ == "__main__":
    main()
