#!/usr/bin/python3

# Author: Adam Gandelman <adamg@canonical.com>
# Copyright (C) 2012 Canonical, Ltd.
# License: GPLv3

# Gather version information of all cloud archive packages from four different
# sources:
#   - Ubuntu release for given OpenStack release (eg, folsom -> quantal)
#   - Cloud Archive Staging PPA
#   - Cloud Archive Proposed repo
#   - Cloud Archive Updates repo
#
# The package set to be tracked is determined by what currently exists in the
# staging PPA.
#
# Optionally include upstream version details, where versions are normalized
# and resulting packages trimmed down to OpenStack deliverables only.
#
# Version info is dumped to JSON files (one for each repo) to a specified
# directory ($HOME/.ca-versions) by default.  These can be used to generate
# a HTML report using the ca-versions.py script.

import common
import git
import json
import os
import re
import sys
import logging
import httplib2
import optparse
import gzip
import shutil
from io import BytesIO
import yaml

from launchpadlib.launchpad import Launchpad

# Map of ubuntu package names to their corresponding upstream project names.
# Most mappings are programatically determined in query_release_deliverables.
# This mapping is just for the tricky names. Used by upstream reports.
PKG_TO_UPSTREAM_MAP = {
    'openstack-trove': 'trove',
    'oslo-sphinx': 'oslosphinx',
    'python-django-openstack-auth': 'django_openstack_auth',
    'python-keystoneauth1': 'keystoneauth',
    'python-os-client-config': 'os-client-config',
    'python-gnocchiclient': 'gnocchiclient',
}

# OpenStack packages that are not tracked in github.com/openstack/releases or
# github.com/openstack/requirements. Used by upstream reports.
PKG_OPENSTACK_NOT_TRACKED = [
    'gnocchi',
    'networking-l2gw',
    'nova-lxd',
    'python-vmware-nsxlib',
    'swift-plugin-s3',
    'vmware-nsx',
]

logging.basicConfig(level=logging.INFO)


def version_from_packages_url(packages, package):
    """Get a package's version."""
    packages = packages.split('\n')
    loc = 0

    def _to_dict(entry):
        out = {}
        for l in entry:
            s = l.split(':')
            if len(s) > 2:
                # Reassemble epics (for clients)
                out[s[0]] = ':'.join(s[1:]).strip()
            elif len(s) >= 2:
                out[s[0]] = ''.join(s[1:]).strip()
        return out

    entries = []
    s = p = None
    while loc < len(packages):
        entry = []
        while packages[loc] != '':
            entry.append(packages[loc])
            loc += 1

        entry = _to_dict(entry)
        entries.append(entry)

        if 'Source' in entry.keys() and entry['Source'] == package:
            s = entry['Version']

        if 'Package' in entry and entry['Package'] == package:
            p = entry['Version']

        loc += 1

    return s or p or None


def versions_from_packages(os_release, pocket, pkgs):
    """Get mapping of package names to versions."""
    url = common.packages_url(os_release, pocket)
    req = httplib2.Http()

    resp, content = req.request(url, "GET")
    if resp.status != 200:
        logging.warn('Could not fetch Packages from {} ({})'.format(
                     url, resp.status))
        return {}

    compressedFile = BytesIO()
    compressedFile.write(content)
    compressedFile.seek(0)
    packages = gzip.GzipFile(fileobj=compressedFile).read()
    packages = packages.decode()
    versions = {}

    for p in pkgs:
        vers = version_from_packages_url(packages, p)
        versions[p] = {}
        versions[p]['original_version'] = vers
        versions[p]['normalized_version'] = vers

    return versions


def init_lp(conn):
    """Initialize launchpad connection."""
    if not conn['lp']:
        logging.info("Initializing connection to LP...")
        conn['lp'] = Launchpad.login_anonymously('openstack', 'production')
    if not conn['ubuntu']:
        conn['ubuntu'] = conn['lp'].distributions['ubuntu']
    return conn['lp']


def query_ca_ppa(conn, ppa='folsom-staging', release='precise'):
    """Get mapping of cloud-archive package names to (version, state)."""
    init_lp(conn)
    ppa = conn['lp'].people['ubuntu-cloud-archive'].getPPAByName(name=ppa)
    distro = ppa.distribution.getSeries(name_or_version=release)
    out = {}
    for pkg in ppa.getPublishedSources(
            distro_series=distro, status='Published'):
        state = 'ok'
        if len(pkg.getPublishedBinaries()) == 0:
            # There are no published binaries, something is wrong with the
            # build. maybe dep wait, ftbfs.
            _st = ' '.join(list(set([b.buildstate for b in pkg.getBuilds()])))
            if 'Failed to build' in _st:
                for log in [b.build_log_url for b in pkg.getBuilds()]:
                    _st += ' <a href="{}">BL</a>'.format(log)
            state = _st

        out[pkg.source_package_name] = {}
        out[pkg.source_package_name]['original_version'] = (
            pkg.source_package_version, state)
        out[pkg.source_package_name]['normalized_version'] = (
            pkg.source_package_version, state)
    return out


def get_ubuntu_pkg_vers(conn, pkg, release):
    """Get package version in Ubuntu based on pocket priority.

    Pocket priority: Security > Updates > Release

    returns: (pocket, version) or (None, None) if pkg doesn't exist in Ubuntu
    rtype: tuple
    """
    init_lp(conn)
    archive, partner = conn['ubuntu'].archives
    release = conn['ubuntu'].getSeries(name_or_version=release)

    def _get_vers(pocket):
        try:
            p = archive.getPublishedSources(exact_match=True,
                                            source_name=pkg,
                                            status='Published',
                                            pocket=pocket,
                                            distro_series=release)[0]
            return p.source_package_version
        except IndexError:
            return ""

    security = _get_vers('Security')
    updates = _get_vers('Updates')

    if security or updates:
        if security > updates:
            return "security", security
        return "updates", updates

    rel = _get_vers('Release')
    if not rel:
        return None, None
    return 'release', rel


def query_ubuntu(conn, pkgs, release):
    """Return dict mapping all pkgs to versions in ubuntu for given release."""
    out = {}
    for p in pkgs:
        vers = get_ubuntu_pkg_vers(conn, p, release)[1]
        out[p] = {}
        out[p]['original_version'] = vers
        out[p]['normalized_version'] = vers
    return out


def normalize_version(version):
    """Convert package version into upstream version format."""
    normalize_regular_expressions = {
        "^\d+:": "",
        "\-.*$": "",
        "\+dfsg.*$": "",
        "\~b1": ".0b1",
        "\~b2": ".0b2",
        "\~b3": ".0b3",
        "\~rc1": ".0rc1",
        "\~rc2": ".0rc2",
        "\~rc3": ".0rc3",
        "\~rc4": ".0rc4",
        "\~cloud.*$": "",
        "ubuntu.*$": "",
    }
    normalized_version = version
    # If 'git' is in version then it's a snapshot and won't match upstream
    # version so no need to normalize.
    if version and 'git' not in version:
        for pattern, replacement in normalize_regular_expressions.items():
            normalized_version = re.sub(r"{}".format(pattern),
                                        replacement,
                                        normalized_version)
    return normalized_version


def _clone_repo(url, out_dir, sub_dir, branch=None):
    """Clone a git repo."""
    dest = os.path.join(out_dir, sub_dir)
    shutil.rmtree(dest, ignore_errors=True)
    if branch:
        git.Repo.clone_from(url, dest, branch=branch)
    else:
        git.Repo.clone_from(url, dest)
    return dest


def query_release_deliverables(pkgs, release, out_dir):
    """Get upstream versions for pkgs from release deliverables."""
    dest = _clone_repo('https://github.com/openstack/releases',
                       out_dir, 'releases')
    release_dir = os.path.join(dest, 'deliverables', release)
    release_files = os.listdir(release_dir)

    branch = 'stable/{}'.format(release)
    try:
        dest = _clone_repo('https://github.com/openstack/requirements',
                           out_dir, 'requirements', branch)
    except git.exc.GitCommandError:
        # if stable branch not found, clone master branch
        branch = 'master'
        dest = _clone_repo('https://github.com/openstack/requirements',
                           out_dir, 'requirements', branch)
    uc_file = os.path.join(dest, 'upper-constraints.txt')

    upstream_versions = {}
    for pkg_name in pkgs:
        # generate upstream project name based on package name
        upstream_name = pkg_name
        if 'client' not in pkg_name:
            upstream_name = re.sub(r'python-', '', pkg_name)
        if pkg_name in PKG_TO_UPSTREAM_MAP.keys():
            upstream_name = PKG_TO_UPSTREAM_MAP[pkg_name]

        upstream_vers = '?'
        release_file = "{}.yaml".format(upstream_name)

        if release_file in release_files:
            # Get latest version from project's release deliverable file.
            # Note: some projects exist in release deliverables but not in
            # upper-constraints.
            stream = open(os.path.join(release_dir, release_file), 'r')
            project_yaml = yaml.load(stream, yaml.FullLoader)
            if 'releases' in project_yaml:
                upstream_vers = (
                    project_yaml['releases'][-1]['version'].split(';')[0]
                )
                # show latest version for extended maintenance branches
                if '-em' in upstream_vers:
                    upstream_vers = '{}-{}'.format(
                        project_yaml['releases'][-2]['version'].split(';')[0],
                        project_yaml['releases'][-1]['version'].split(';')[0]
                    )
            else:
                upstream_vers = 'None'

        # If version exists in upper-constraints, use that instead. Release
        # deliverables files can get updated before upper-constraints.
        stream = open(uc_file, 'r')
        for line in stream:
            lline = line.lower().rstrip()
            if upstream_name == re.sub(r'===.*$', '', lline):
                upstream_vers = re.sub(r'^.*===', '', lline)

        # If version isn't tracked in release deliverable file or in upper-
        # contraints, then don't report it *unless* it really is an OpenStack
        # package that is just not tracked.
        if upstream_vers == '?':
            if pkg_name in PKG_OPENSTACK_NOT_TRACKED:
                upstream_vers = 'not-tracked'
            else:
                continue

        upstream_versions[pkg_name] = {}
        upstream_versions[pkg_name]['original_version'] = (
            upstream_vers.split(';')[0])
        upstream_versions[pkg_name]['normalized_version'] = (
            upstream_vers.split(';')[0])

    return upstream_versions


def trim_packages_file(filename, packages):
    """Trim file to include only packages that are in packages dict."""
    f = open(filename, 'r')
    pkgs = json.loads(f.read())

    keys = list(pkgs.keys())
    for k in keys:
        if k not in packages.keys():
            del pkgs[k]

    f = open(filename, 'w')
    f.write(json.dumps(pkgs))


def main():
    usage = ('usage: {}prog [options] '.format(
             str(sorted(common.VERSION_MAP.keys()))))
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-o', '--output_dir', action='store', dest='out_dir',
                      default=os.path.join(os.getenv('HOME'), '.ca-versions'),
                      help='Directory to dump JSON '
                           '(Default: $HOME/.ca-versions)')
    parser.add_option('-u', '--upstream', dest='include_upstream',
                      action='store_true', default=False,
                      help='Get upstream versions and normalize/trim results.')
    (opts, args) = parser.parse_args()

    if len(args) == 0:
        logging.error("must provide openstack release codename (eg, folsom).")
        sys.exit(1)

    os_release = args.pop()
    if os_release not in common.VERSION_MAP.keys():
        logging.error("Invalid openstack release ({}).".format(os_release))
        logging.error("Expect one of: {}".format(
                      str(common.VERSION_MAP.keys())))
        sys.exit(1)

    if not os.path.exists(opts.out_dir):
        os.mkdir(opts.out_dir)

    conn = {}
    conn["lp"] = None
    conn["ubuntu"] = None
    stable_rel = "precise"

    stable_rel = common.VERSION_MAP[os_release]['ubuntu_stable']
    dev_rel = common.VERSION_MAP[os_release]['ubuntu_dev']

    # Gather versions from the staging ppa for this release.
    # the packages that have been pushed here will determine what is tracked
    # in other places.
    logging.info("Querying package list and versions from staging PPA.")
    if 'staging_ppa' in common.VERSION_MAP[os_release]:
        staging_ppa = common.VERSION_MAP[os_release]['staging_ppa']
    else:
        staging_ppa = '{}-staging'.format(os_release)

    pockets = common.VERSION_MAP[os_release].get('ca_pockets',
                                                 ['proposed', 'updates'])

    staging_versions = query_ca_ppa(conn, ppa=staging_ppa,
                                    release=stable_rel)
    f = open(os.path.join(opts.out_dir,
             '{}_staging_versions'.format(os_release)), 'w+')
    if opts.include_upstream:
        staging_versions_keys = staging_versions.keys()
        for pkg_name in staging_versions_keys:
            orig_version = staging_versions[pkg_name]['original_version'][0]
            orig_state = staging_versions[pkg_name]['original_version'][1]
            staging_versions[pkg_name]['normalized_version'] = (
                normalize_version(orig_version), orig_state)
    f.write(json.dumps(staging_versions))
    f.close()

    # With the names of the packages published to the staging ppa, we can
    # determine what we need to track in ubuntu.
    logging.info("Querying Ubuntu versions for all packages.")
    ubuntu_versions = query_ubuntu(conn, staging_versions.keys(), dev_rel)
    f = open(os.path.join(opts.out_dir,
             "{}_ubuntu_versions".format(os_release)), 'w+')
    if opts.include_upstream:
        ubuntu_versions_items = ubuntu_versions.items()
        for pkg_name, version in ubuntu_versions_items:
            ubuntu_versions[pkg_name]['normalized_version'] = (
                normalize_version(version['original_version']))
    f.write(json.dumps(ubuntu_versions))
    f.close()

    # And then we gather the proposed and updates package versions
    for p in pockets:
        logging.info("Scraping Packages list for CA pocket: {}".format(p))
        versions = versions_from_packages(os_release, p,
                                          staging_versions.keys())
        out = os.path.join(opts.out_dir,
                           '{}_{}_versions'.format(os_release, p))
        f = open(out, 'w+')
        if opts.include_upstream:
            versions_items = versions.items()
            for pkg_name, version in versions_items:
                if version['original_version']:
                    version['normalized_version'] = (
                        normalize_version(version['original_version']))
        f.write(json.dumps(versions))

    # Finally gather the upstream versions if requested
    if opts.include_upstream:
        # With the names of the packages published to the staging ppa, we can
        # determine the upstream versions we need to track.
        logging.info("Querying upstream versions for packages "
                     "from release deliverables repo.")
        upstream_versions = query_release_deliverables(
            staging_versions, os_release, opts.out_dir)

        # Trim the results to only include packages that we've detected
        # upstream versions for.
        for k in upstream_versions.keys():
            if upstream_versions[k] is None:
                del upstream_versions[k]

        out = os.path.join(opts.out_dir,
                           '{}_upstream_versions'.format(os_release))
        f = open(out, 'w')
        f.write(json.dumps(upstream_versions))

        out = os.path.join(opts.out_dir,
                           '{}_staging_versions'.format(os_release))
        trim_packages_file(out, upstream_versions)

        out = os.path.join(opts.out_dir,
                           '{}_ubuntu_versions'.format(os_release))
        trim_packages_file(out, upstream_versions)

        for p in pockets:
            out = os.path.join(opts.out_dir,
                               '{}_{}_versions'.format(os_release, p))
            trim_packages_file(out, upstream_versions)

if __name__ == "__main__":
    main()
