from datetime import datetime, timedelta
from launchpadlib.launchpad import Launchpad
try:
    from lazr.restfulclient.errors import Unauthorized
except ImportError:
    # Old versions of lplib doesn't have it 
    pass


def calculateDays(date):
    if date != "-":
        t = datetime.utcnow()
        try:
            now = datetime.strptime(date.split(".")[0], "%Y-%m-%d %H:%M:%S")
        except AttributeError:
            # date is a datetime object already
            now = date
            t = t.replace(tzinfo=now.tzinfo)
        age = t - now
        days = age.days
    else:
        days = 0
    return int(days)

def calculateHours(date):
    if date != "-":
        t = datetime.utcnow()
        try:
            now = datetime.strptime(date.split(".")[0], "%Y-%m-%d %H:%M:%S")
        except AttributeError:
            # date is a datetime object already
            now = date
            t = t.replace(tzinfo=now.tzinfo)
        age = t - now
        seconds = int(age.seconds)
        hours = seconds/3600
    else:
        hours = 0
    return int(hours)

def calculateMinutes(date):
    if date != "-":
        t = datetime.utcnow()
        try:
            now = datetime.strptime(date.split(".")[0], "%Y-%m-%d %H:%M:%S")
        except AttributeError:
            # date is a datetime object already
            now = date
            t = t.replace(tzinfo=now.tzinfo)
        age = t - now
        seconds = int(age.seconds)
        minutes = seconds/60
    else:
        minutes = 0
    return int(minutes)



class Bug:

    def __init__(self, lp_bug):
        if lp_bug is not None:
            self.id = lp_bug.id
            self.linked_branches = lp_bug.linked_branches
            self.attachments = lp_bug.attachments
            self.title = lp_bug.title
            self.tags = lp_bug.tags
            self.activity = lp_bug.activity
            try:
                self.last_activity = self.activity[len(self.activity)-1]
                self.last_changed = self.last_activity.datechanged
                self.whatchanged = self.last_activity.whatchanged
                self.whochanged = self.last_activity.person.name
            except:
                self.last_changed = lp_bug.date_last_updated

            self.last_touched_days = calculateDays(self.last_changed)
            self.last_touched_hours = calculateHours(self.last_changed)
            self.last_touched_minutes = calculateMinutes(self.last_changed)

        self.bug = lp_bug

    def get_attachments(self):
        attachments = self.attachments
        if attachments:
            patches = [att.title for att in attachments if att.type == "Patch"]
        else:
            attachments = []
            patches = []
        return {'attachments': attachments,
                'patches': patches}

    def get_linked_branches(self):
        branches = []
        linked_branches = self.linked_branches
        try:
            try:
                if linked_branches:
                    for branch in linked_branches:
                        branches.append("lp:" + branch.branch.unique_name)
            except Unauthorized, exc:
                #print "(private branch, skipping)"
                pass
        except NameError:
            # Some older versions of lplib doesn't have Unauthorized.
            pass

    def has_reviewer_subscribed(bug, reviewer_team_name="ubuntu-sru"):
        subscriptions = bug.subscriptions
        if subscriptions:
            all_subscriptions = [subscription.person.name
                for subscription in subscriptions]
            for subscription_name in all_subscriptions:
                if subscription_name == reviewer_team_name:
                    return True
        return False


class BugTask:

    def __init__(self, lp_bugtask):
        if lp_bugtask is not None:
            self.bug_id = lp_bugtask.bug.id
            self.assignee = lp_bugtask.assignee
            self.status = lp_bugtask.status
            self.importance = lp_bugtask.importance
            self.bug = Bug(lp_bugtask.bug)
            self.date_created = lp_bugtask.date_created
            self.age_days = calculateDays(self.date_created)
            self.age_hours = calculateHours(self.date_created) % 24
            self.target = lp_bugtask.target.name
            self.milestone = lp_bugtask.milestone
        self.bugtask = lp_bugtask

    def has_assignee(self):
        return (self.assignee is not None)

    def check_bugtask(self):
        current_bug = self.bug
        print "[Bug:%s] %s (%s)" % (current_bug.id, current_bug.title,
            self.status)
        attachments = current_bug.get_attachments()
        linked_branches = current_bug.get_linked_branches()
        current_bug.has_reviewer_subscribed()


class Package:

    def __init__(self, pkg, name=None, owner=None):
        if pkg is not None:
            self.pkg = pkg
            self.name = pkg.name
            self.owner = owner
            self.bugtasks = []

    def get_bugtasks(self, importance=None, status=None, subscriber=None):
        if status is not None and importance is not None:
            self.bugtasks = self.pkg.searchTasks(order_by="-datecreated",
                status=status, importance=importance,
                bug_subscriber=subscriber)
        elif status is not None:
            self.bugtasks = self.pkg.searchTasks(order_by="-datecreated",
                status=status, bug_subscriber=subscriber)
        elif importance is not None:
            self.bugtasks = self.pkg.searchTasks(order_by="-datecreated",
                importance=importance, bug_subscriber=subscriber)
        self.bugtasks = self.pkg.searchTasks(order_by="-datecreated",
                bug_subscriber=subscriber)
        return len(self.bugtasks)


class LaunchpadUtils:

    def __init__(self, script_name, lp_instance):
        self.script_name = script_name
        self.lp_instance = lp_instance

    def login(self):
        try:
            self.lp = Launchpad.login_with(self.script_name, self.lp_instance)
        except Exception, exc:
            raise

    def get_team(self, team_name):
        try:
            team = self.lp.people[team_name]
            if team is None:
                # XXX: Better error handling here!
                # Raise an exception
                print "Team not found"
        except:
            # Error fetching team.
            raise

        return Team(team.name, team)


class Team:

    def __init__(self, name, lp_team):
        self.name = name
        self.lp_team = lp_team
        self.subscribed_packages = []
        self.bugtasks = {}
        self.bugtasks_importance = {}
        self.bugtasks_status = {}

    def get_subscribed_packages(self):
        try:
            pkgs = self.lp_team.getBugSubscriberPackages()
            self.subscribed_packages = [Package(pkg) for pkg in pkgs]
        except:
            # Error fetching packages
            raise
        return len(self.subscribed_packages)

    def get_bugtasks_for_pkg(self, pkg, importance=None, status=None,
            subscriber=None):
        pkg.get_bugtasks(importance, status, subscriber)
        for bugtask in pkg.bugtasks:
            date_created = bugtask.date_created.strftime("%Y-%m-%d")
            self.bugtasks.setdefault(date_created, [])
            bt = BugTask(bugtask)
            self.bugtasks[date_created].append(bt)
            self.bugtasks_importance.setdefault(bt.importance, [])
            self.bugtasks_importance[bt.importance].append(bt)
            self.bugtasks_status.setdefault(bt.status, [])
            self.bugtasks_status[bt.status].append(bt)



class Distro:

    def __init__(self, lp_distro):
        if lp_distro is not None:
            self.name = lp_distro.name
            self.lp_distro = lp_distro
            self.series = {}
        self.bugtasks = {}
        self.bugtasks_importance = {}
        self.bugtasks_status = {}

    def add_series(self, distro_series_name, distro_series_version):
        try:
            new_series = self.series[distro_series_name]
            raise
            # We already have a series with this name. Raise a proper
            # Exception here
        except KeyError:
            # No series yet! I can create one.
            new_series = DistroSeries(distro_series_name,
                distro_series_version)
            self.series.setdefault(distro_series_name, None)
            self.series[distro_series_name] = new_series

    def get_bugtasks_per_bug_subscriber(self, lp_person, importance=None,
            status=None):
        return self.lp_distro.searchTasks(bug_subscriber=lp_person,
            importance=importance, status=status)

    def get_bugtasks_per_bug_supervisor(self, bug_supervisor, importance=None,
            status=None, bug_subscriber=None):
        return self.lp_distro.searchTasks(bug_supervisor=bug_supervisor,
            importance=importance, status=status, tags=["-notserv"],
            bug_subscriber=bug_subscriber)

    def get_bugtasks(self, bug_supervisor, importance=None, status=None,
            bug_subscriber=None):
        bugtasks = self.get_bugtasks_per_bug_supervisor(bug_supervisor,
            importance, status, bug_subscriber)
        for bugtask in bugtasks:
            date_created = bugtask.date_created.strftime("%Y-%m-%d")
            self.bugtasks.setdefault(date_created, [])
            bt = BugTask(bugtask)
            self.bugtasks[date_created].append(bt)
            if importance is not None and len(importance) == 1:
                self.bugtasks_importance.setdefault(importance[0], [])
                self.bugtasks_importance[importance[0]].append(bt)
            else:
                self.bugtasks_importance.setdefault(bt.importance, [])
                self.bugtasks_importance[bt.importance].append(bt)
            if status is not None and len(status) == 1:
                self.bugtasks_status.setdefault(status[0], [])
                self.bugtasks_status[status[0]].append(bt)
            else:
                self.bugtasks_status.setdefault(bt.status, [])
                self.bugtasks_status[bt.status].append(bt)

    def get_last_touched_bugtasks(self, bug_supervisor, importance=None,
            status=None, bug_subscriber=None, time_in_minutes=120):
        minutes_ago = datetime.utcnow() - timedelta(minutes=time_in_minutes)
        return [BugTask(bt) for bt in self.lp_distro.searchTasks(
            bug_supervisor=bug_supervisor, importance=importance,
            status=status, tags=["-notserv"], modified_since=minutes_ago,
            omit_duplicates=False)]


class DistroSeries:

    def __init__(self, distro_series_name, distro_series_version):
        self.name = distro_series_name
        self.version = distro_series_version
        self.development = False
        self._type = "Regular"  # Or LTS

    def is_current(self):
        pass
        # check in launchpad
