#!/usr/bin/python
# Author: Brian Murray <brian@canonical.com>
# Copyright (C) 2011 Canonical, Ltd.
# License: GPLv3

import datetime
import optparse
import sys

from cgi import escape
from lpltk import LaunchpadService


def html_report(team_name, tasks, target_date):
    datafile = open('%s-old-in-progress-bug-tasks.html' % team_name, 'w')
    supported_releases = ['Dapper', 'Hardy', 'Lucid', 'Maverick', 'Natty']
    importances = {'Critical': 1, 'High': 2, 'Medium': 3, 'Low': 4,
        'Wishlist': 5, 'Undecided': 6}
    statuses = {"Won't Fix": 1, 'New': 2, 'Incomplete': 3, 'Confirmed': 4,
        'Triaged': 5, 'In Progress': 6, 'Fix Committed': 7, 'Fix Released': 8}
    templatefile = open('template.html', 'r')
    counts = {}

    for line in templatefile:
        if line.strip() == "<!-- *** Title Space *** -->":
            datafile.write("Old assigned In Progress bug tasks for the %s team\n" % team_name)
        elif line.strip() == "<!-- *** Header Space *** -->":
            datafile.write("Old assigned In Progress  bug tasks for the %s team\n" % team_name)
        elif line.strip() == "<!-- *** Paragraph Space *** -->":
            datafile.write("In Progress bug tasks older than %s and assinged to %s members.  Some rows will be a different color, this indicates they are tagged as a regression.  All of the column headers are sortable give them a click!" % (opt.days, team_name))
        elif line.strip() == "<!-- *** Table Header Space *** -->":
            table_header = "<th>#</th>"  # bug number in this column
            table_header += "<th>Summary</th>"
            table_header += "<th>In</th>"
            table_header += "<th>Importance</th>"
            table_header += "<th>Status</th>"
            table_header += "<th>Release</th>"
            table_header += "<th>Assignee</th>"
            table_header += "<th>Date In Progress</th>"
            table_header += "<th>Regression</th>"
            table_header += "\n"
            datafile.write(table_header)
        elif line.strip() == "<!-- *** Table Body Space *** -->":

            for task in tasks:
                #started = task.date_in_progress
                #if started is None:
                #    continue
                #if started.date() > target_date.date():
                #    continue
                if [tag for tag in task.bug.tags if 'regression-' in tag]:
                    # change the class to highlight for any bug that has a regression- tag in it
                    table_row = '<tr class="highlight">\n'
                else:
                    table_row = '<tr>\n'
                table_row += '<td id="bug_number">%s</td>\n' % task.bug.id
                table_row += "<td><a href='http://launchpad.net/bugs/%s'>%s</a></td>\n" % (task.bug.id, escape(task.bug.title.encode('utf-8')))
                table_row += '<td id="target">%s</td>\n' % task.bug_target_display_name.encode('utf-8')
                importance = task.importance.encode('utf-8')
                table_row += '<td id="importance"><span class="sortkey">%s</span>%s</td>\n' % ( importances[importance], importance )
                status = task.status.encode('utf-8')
                table_row += '<td id="status"><span class="sortkey">%s</span>%s</td>\n' % ( statuses[status], status )

                table_row += '<td id="release">'
                affected_releases = []
                for release in supported_releases:
                    rtasks = [task.bug_target_display_name for task in task.related_tasks]
                    if release in rtasks:
                        affected_releases.append('%s%s' % (release[0], release[0]))
                if affected_releases:
                    table_row += ', '.join(affected_releases)
                else:
                    table_row += ''
                table_row += '</td>\n'

                if task.assignee is not None:
                    assignee = task.assignee.display_name
                    try:
                        table_row += '<td id="assignee">%s</td>\n' % task.assignee.lpperson.display_name.encode('utf-8')
                    except UnicodeDecodeError:
                        table_row += '<td id="assignee">%s</td>\n' % task.assignee.name
                else:
                    table_row += '<td id="assignee"></td>\n'
                #table_row += "<td></td>"
                if task.date_in_progress is not None:
                    try:
                        table_row += '<td id="date_in_progress">%s</td>\n' % task.date_in_progress[:10].encode('utf-8')
                    except TypeError:
                        table_row += '<td id="date_in_progress">%s</td>\n' % task.date_in_progress.strftime('%Y-%m-%d')
                else:
                    table_row += '<td id="date_in_progress"></td>\n'
                items = [tag for tag in task.bug.tags if 'regression-' in tag]
                if len(items) > 0:
                    table_row += '<td id="regression">'
                    for item in items:
                        table_row += "%s " % item.encode('utf-8')
                    table_row += '</td>\n'
                else:
                    table_row += '<td id="regression"></td>\n'
                table_row += "</tr>"
                table_row += "\n"
                datafile.write(table_row)
                # only count rows that show up on the report not hokey ones
                counts['total'] = counts.setdefault('total', 0) + 1
        elif line.strip() == "<!-- *** Last Paragraph Space *** -->":
            try:
                datafile.write("A total of %s bug tasks have been assigned and In Progress since %s.<br><br>" % (counts['total'], target_date.date()))
            except KeyError:
                datafile.write("The %s team has no assigned In Progress bug tasks older than %s days.<br><br>" % (team_name, opt.days))
            datafile.write('<br>')
            datafile.write('Is there a problem with this report? File a bug about the <a href="https://bugs.launchpad.net/ubuntu-qa-tools/+filebug">ubuntu-qa-tools.</a><br>')
            datafile.write('<br>')
            now = datetime.datetime.now()
            datafile.write('Report generated at: %s<br>' % now.strftime('%Y-%m-%d %H:%M'))
        else:
            datafile.write(line)


def find_team_members(team):
    team_members = []
    for member in team.members:
        if member.is_team:
            continue
        team_members.append(member)
    return team_members

if __name__ == '__main__':
    lp = LaunchpadService(config={})

    ubuntu = lp.distributions['ubuntu']
    parser = optparse.OptionParser()

    parser.add_option("--days", help="Filter on bugs In Progress longer than \
        X day(s)")
    parser.add_option("--team", help="Filter on bugs assigned to members of \
        team")
    parser.add_option("--html", help="Output to html", action="store_true")

    (opt, args) = parser.parse_args()
    team_mapping = {
        'canonical-foundations': 'foundations-bugs',
        'canonical-server': 'ubuntu-server',
        'canonical-desktop-team': 'desktop-packages'}
    team = lp.launchpad.people[opt.team]
    bug_team = lp.launchpad.people[team_mapping[opt.team]]
    team_members = find_team_members(team)
    today = datetime.datetime.utcnow()
    target_date = today - datetime.timedelta(int(opt.days))

    collected_tasks = []

    for team_member in team_members:
        tasks = ubuntu.search_tasks(bug_supervisor=bug_team,
            assignee=team_member, status=['In Progress'])


        for task in tasks:
            started = task.date_in_progress
            if started is None:
                continue
            if started.date() < target_date.date():
                collected_tasks.append(task)

    if opt.html:
        html_report(opt.team, collected_tasks, target_date)
        sys.exit(0)

    for task in collected_tasks:
        print 'LP: #%s %s In Progress since %s' % \
            (task.bug.id, task.bug.title, started.date())
