#!/usr/bin/python
# Author: Brian Murray <brian@canonical.com>
# Copyright (C) 2010 Canonical, Ltd.
# License: GPLv3
#
# Given a team name retrieve json data files from reports.qa.ubuntu.com with
# bug information for that team, additionally query Launchpad for bug tasks
# about packages to which the team is subscribed with specific tags.  The goal
# is to provide a plain text notification of bugs to review.
#
# To avoid duplicate notifications, run once a week as timedelta is set to 7.

from lpltk import LaunchpadService
from urllib2 import urlopen

import datetime
import optparse
import simplejson as json
import sys


def get_json_data(team, source):
    datafile = urlopen(
        'http://reports.qa.ubuntu.com/reports/%s/%s-%s.json' %
        (team, team, source))
    json_data = json.load(datafile)
    return json_data


def get_arsenal_json_data(team, source):
    data = []
    url = 'http://reports.qa.ubuntu.com/reports/%s/%s-%s.json' % (team, team, source)
    datafile = urlopen(url)
    json_data = json.load(datafile)
    for task in json_data['tasks']:
        if not json_data['tasks'][task]:
            continue
        task_info = {}
        task_info['bug'] = task
        task_info['title'] = json_data['tasks'][task][0]['bug']['title']
        data.append(task_info)
    return data


if __name__ == '__main__':
    launchpad = LaunchpadService(config={'read_only': False,
                                 'display_name': 'defect-analysts-spike'})

    parser = optparse.OptionParser(usage="usage: %prog --team TEAM")
    parser.add_option('--team', help='Launchpad team to check')
    (opt, args) = parser.parse_args()

    ubuntu = launchpad.distributions['ubuntu']

    if opt.team:
        team_name = opt.team
        team = launchpad.launchpad.people[team_name]
    else:
        print('A Launchpad team name must be provided.')
        sys.exit(1)

    today = datetime.datetime.utcnow()
    target_date = today - datetime.timedelta(int(7))

    # search for certain types of bugs rather than using a json file
    # NB: can't use an lpltk person object to search tasks
    for tag in ['rls-x-incoming', 'rls-bb-incoming', 'rls-dd-incoming',
                'rls-ee-incoming', 'iso-testing', 'regression-update',
                'regression-proposed', 'regression-release']:
        print("%s" % tag)
        print("-" * len(tag))
        for task in ubuntu.search_tasks(modified_since=target_date,
                                        bug_supervisor=team, tags=tag):
            bug = task.bug
            try:
                if bug.date_tag_added(tag).replace(tzinfo=None) < target_date:
                    continue
            except AttributeError:
                pass
            print("LP: #%s (%s)" % (bug.id, bug.title))

    for source in ['recent-duplicate-bugs', 'resolved-upstream-bugs',
                   'recent-triaged', 'recent-high', 'recent-confirmed',
                   'reopened', 'nominated-bugs', 'hottest-bugs']:
        print("%s" % source)
        print("-" * len(source))
        if source in ['recent-triaged', 'recent-high', 'recent-confirmed',
                      'reopened']:
            # if this is to be run daily then these should be filtered on the
            # date the action happened since the json files use within the
            # past seven days, this is hard for recent-high though as
            # date_importance_set does not exist in the json file
            json_data = get_arsenal_json_data(team_name, source)
        else:
            json_data = get_json_data(team_name, source)

        for data in json_data:
            if source == 'recent-duplicate-bugs':
                if data['recent_dupes'] == 1:
                    print("LP: #%s (%s) received %s new dupe out \
of a total of %s dupes" %
                          (data['master'], data['title'], data['recent_dupes'],
                           data['dupe_count']))
                else:
                    print("LP: #%s (%s) received %s new dupes out \
of a total of %s dupes" %
                          (data['master'], data['title'], data['recent_dupes'],
                           data['dupe_count']))
            elif source == 'resolved-upstream-bugs':
                for bugwatch in data['watches']:
                    update = data['watches'][bugwatch]
                    try:
                        date_updated = datetime.datetime.strptime(
                            update.split()[0], '%Y-%m-%d')
                    except ValueError:
                        continue
                    if date_updated >= target_date:
                        print("LP: #%s (%s)" % (data['bug'], data['title']))
            elif source == 'nominated-bugs':
                print("  %s" % data)
                print("  " + "-" * len(data))
                for bug in json_data[data]:
                    print("LP: #%s (%s)" % (bug['bug'], bug['title']))
            else:
                print("LP: #%s (%s)" % (data['bug'], data['title']))
