#!/usr/bin/python
# Author: Brian Murray <brian@canonical.com>
# Copyright (C) 2013 Canonical, Ltd.
# License: GPLv3
#
# Find bug tasks for packages to which a team is subscribed that are nominated
# for the Current Stable or Supported releases of Ubuntu

import optparse
import simplejson as json

import datetime
from lpltk import LaunchpadService


def find_nominations(release, target_date):
    bugs = {}
    nominated_tasks = [task for task in \
                       team.searchTasks(nominated_for=release)]

    nominations = {}
    for task in nominated_tasks:
        nomination = task.bug.getNominationFor(target=release)
        if str(nomination) not in str(nominations):
            nom_date = nomination.date_created.date()
            if nom_date > target_date.date():
                nominations[nomination] = nom_date
                bug_num = task.bug.id
                bugs[bug_num] = {}
                bugs[bug_num]['title'] = task.bug.title
                bugs[bug_num]['date_nominated'] = \
                    datetime.datetime.strftime(nom_date, "%Y-%m-%d")
    return bugs


if __name__ == '__main__':
    lp = LaunchpadService(config=
        {'read_only': False,
         'display_name': 'foundations-recent-package-bug-tasks'})

    ubuntu = lp.load_project('ubuntu')

    parser = optparse.OptionParser()
    parser.add_option("--team", help="Team to search for")
    parser.add_option("--within", help="Filter on bugs reported within \
X day(s)")

    (opt, args) = parser.parse_args()

    if opt.team:
        team = lp.launchpad.people[opt.team]

    releases = [s for s in ubuntu.series if s.status in
                ['Current Stable Release', 'Supported', 'Active Development',
                 'Pre-release Freeze']]

    today = datetime.datetime.utcnow()
    target_date = today - datetime.timedelta(int(opt.within))

    data = {}
    json_filename = "%s-nominated-bugs.json" % (opt.team)

    for release in releases:
        rname = release.name
        data[rname] = []
        bugs = find_nominations(release, target_date)
        for bug in bugs:
            data[rname].append({'bug': bug,
                'title': bugs[bug]['title'],
                'date_nominated': bugs[bug]['date_nominated']})

    json_file = open(json_filename, 'w')
    json_file.write(json.dumps(data, indent=4))
    json_file.close()
